<?php
namespace Magenest\Salesforce\Controller\Adminhtml\Mass;

use Magenest\Salesforce\Controller\Adminhtml\Queue\AbstractionAction;
use Magenest\Salesforce\Model\Queue;
use Magenest\Salesforce\Model\ResourceModel\Queue as QueueResource;
use Magento\Backend\App\Action\Context;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Customer\Api\CustomerRepositoryInterface;
use Magento\Framework\Controller\ResultFactory;
use Magento\Framework\Exception\LocalizedException;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Ui\Component\MassAction\Filter;
use Psr\Log\LoggerInterface;

class AddToQueue extends AbstractionAction
{
    /**
     * Authorization level of a basic admin session
     *
     * @see _isAllowed()
     */
    const ADMIN_RESOURCE = 'Magenest_Salesforce::queue';

    /** @var QueueResource */
    protected $_queueResource;

    /**
     * @var string
     */
    protected $redirectUrl = '*/*/index';

    /** @var Filter */
    protected $_filter;

    /** @var LoggerInterface */
    protected $_logger;

    /**
     * @var array
     */
    protected $_salesforceObject = [
        'customer' => [
            Queue::TYPE_LEAD,
            Queue::TYPE_CONTACT,
            Queue::TYPE_ACCOUNT
        ],
        'order'    => [
            Queue::TYPE_ORDER,
            Queue::TYPE_OPPORTUNITY
        ],
        'product'  => [
            Queue::TYPE_PRODUCT
        ],
        'campaign' => [
            Queue::TYPE_CAMPAIGN
        ]
    ];

    /**
     * AddToQueue constructor.
     *
     * @param QueueResource $queueResource
     * @param Filter $filter
     * @param Context $context
     * @param CustomerRepositoryInterface $customerRepository
     * @param ProductRepositoryInterface $productRepository
     * @param OrderRepositoryInterface $orderRepository
     * @param \Magento\Framework\App\ResourceConnection $resourceConnection
     * @param \Magento\Framework\MessageQueue\PublisherInterface $publisher
     * @param \Magenest\Salesforce\Logger\Logger $logger
     * @param \Magenest\Salesforce\Helper\Data $helperData
     * @param QueueResource\CollectionFactory $queueCollectionFactory
     * @param \Magento\Newsletter\Model\ResourceModel\Subscriber\CollectionFactory $subscriberCollectionFactory
     * @param \Magento\Framework\View\Result\PageFactory $resultPageFactory
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Config\Model\ResourceModel\Config $config
     * @param \Magento\Customer\Model\ResourceModel\Customer\CollectionFactory $customerCollection
     * @param \Magento\Framework\Serialize\Serializer\Json $json
     * @param \Magento\CatalogRule\Model\ResourceModel\Rule\CollectionFactory $ruleColFactory
     * @param \Magento\Sales\Model\ResourceModel\Order\CollectionFactory $orderColFactory
     * @param \Magento\Framework\Api\SearchCriteriaBuilder $searchCriteriaBuilder
     * @param \Magento\Catalog\Model\ResourceModel\Product\CollectionFactory $productColFactory
     * @param \Magento\Framework\Api\FilterBuilder $filterBuilder
     */
    public function __construct(
        QueueResource                                                        $queueResource,
        Filter                                                               $filter,
        Context                                                              $context,
        CustomerRepositoryInterface                                          $customerRepository,
        ProductRepositoryInterface                                           $productRepository,
        OrderRepositoryInterface                                             $orderRepository,
        \Magento\Framework\App\ResourceConnection                            $resourceConnection,
        \Magento\Framework\MessageQueue\PublisherInterface                   $publisher,
        \Magenest\Salesforce\Logger\Logger                                   $logger,
        \Magenest\Salesforce\Helper\Data                                     $helperData,
        \Magenest\Salesforce\Model\ResourceModel\Queue\CollectionFactory     $queueCollectionFactory,
        \Magento\Newsletter\Model\ResourceModel\Subscriber\CollectionFactory $subscriberCollectionFactory,
        \Magento\Framework\View\Result\PageFactory                           $resultPageFactory,
        \Magento\Framework\App\Config\ScopeConfigInterface                   $scopeConfig,
        \Magento\Config\Model\ResourceModel\Config                           $config,
        \Magento\Customer\Model\ResourceModel\Customer\CollectionFactory     $customerCollection,
        \Magento\Framework\Serialize\Serializer\Json                         $json,
        \Magento\CatalogRule\Model\ResourceModel\Rule\CollectionFactory      $ruleColFactory,
        \Magento\Sales\Model\ResourceModel\Order\CollectionFactory           $orderColFactory,
        \Magento\Framework\Api\SearchCriteriaBuilder                         $searchCriteriaBuilder,
        \Magento\Catalog\Model\ResourceModel\Product\CollectionFactory       $productColFactory,
        \Magento\Framework\Api\FilterBuilder                                 $filterBuilder
    ) {
        $this->_queueResource = $queueResource;
        $this->_filter        = $filter;
        $this->_logger        = $logger;
        parent::__construct(
            $resourceConnection,
            $publisher,
            $logger,
            $helperData,
            $queueResource,
            $queueCollectionFactory,
            $subscriberCollectionFactory,
            $resultPageFactory,
            $scopeConfig,
            $config,
            $customerCollection,
            $json,
            $ruleColFactory,
            $orderColFactory,
            $orderRepository,
            $searchCriteriaBuilder,
            $productColFactory,
            $filterBuilder,
            $customerRepository,
            $context
        );
    }

    /**
     * @return \Magento\Backend\Model\View\Result\Redirect|\Magento\Framework\App\ResponseInterface|\Magento\Framework\Controller\ResultInterface
     */
    public function execute()
    {
        /** @var \Magento\Backend\Model\View\Result\Redirect $resultRedirect */
        $resultRedirect = $this->resultFactory->create(ResultFactory::TYPE_REDIRECT);
        try {
            $params = $this->getRequest()->getParams();
            $type   = $params['type'] ?? null;
            if (!isset($params['type'])) {
                throw new LocalizedException(__("Object is not determined"));
            }
            if (isset($this->_salesforceObject[$type])) {
                $salesforceObject = $this->_salesforceObject[$type];
                $selectedParam    = $this->getAllIds($type, $this->getRequest()->getParam(Filter::SELECTED_PARAM));
                foreach ($salesforceObject as $object) {
                    $queueIds = $this->getQueueIdsByType($object);
                    $this->_queueResource->deleteMultiQueues($queueIds);
                    $this->_queueResource->publishData($selectedParam, $object);
                }
                $this->messageManager->addSuccessMessage(
                    __('A total of %1 data have been added to queue.', count($selectedParam))
                );
                if ($type == 'customer') {
                    return $resultRedirect->setPath('customer/index/index');
                } elseif ($type == 'order') {
                    return $resultRedirect->setPath('sales/order/index');
                } elseif ($type == 'product') {
                    return $resultRedirect->setPath('catalog/product/index');
                }
            }
        } catch (\Throwable $exception) {
            $this->_logger->debug($exception->getMessage());
            $this->messageManager->addErrorMessage($exception->getMessage());
        }

        return $resultRedirect->setPath($this->redirectUrl);
    }

    /**
     * @param $type
     * @param $filter
     * @return array
     */
    public function getAllIds($type, $filter)
    {
        switch ($type) {
            case 'customer':
                $collection = $this->customerCollection->create()->addFieldToSelect('entity_id');
                break;
            case 'order':
                $collection = $this->_orderColFactory->create();
                break;
            case 'product':
                $collection = $this->_productColFactory->create()->addFieldToSelect('entity_id');
                break;
            default:
                return [];
        }

        if (!empty($filter)) {
            $collection->addFieldToFilter('entity_id', ['in' => $filter]);
        }

        if ($type == 'order') {
            return $collection->getColumnValues(OrderInterface::INCREMENT_ID);
        } else {
            return $collection->getAllIds();
        }
    }
}
