<?php
namespace Magenest\Salesforce\Model\Sync;

use Magenest\Salesforce\Logger\Logger;
use Magenest\Salesforce\Model\Connector;
use Magenest\Salesforce\Model\Data;
use Magenest\Salesforce\Model\ReportFactory as ReportFactory;
use Magenest\Salesforce\Model\RequestLogFactory;
use Magenest\Salesforce\Model\ResourceModel\Queue as QueueResource;
use Magenest\Salesforce\Model\ResourceModel\Queue\CollectionFactory as QueueCollectionFactory;
use Magenest\Salesforce\Model\ResourceModel\Report\CollectionFactory as ReportCollectionFactory;
use Magento\Config\Model\ResourceModel\Config as ResourceModelConfig;
use Magento\Framework\App\Config\ScopeConfigInterface as ScopeConfigInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\HTTP\ClientFactory;
use Magento\Framework\Model\ResourceModel\Iterator;
use Magento\Framework\Serialize\Serializer\Json;
use Magento\Framework\Serialize\Serializer\Serialize;
use Magento\Sales\Model\Order as OrderModel;
use Magento\Sales\Model\OrderFactory;

class Opportunity extends Connector
{
    const SALESFORCE_OPPORTUNITY_ATTRIBUTE_CODE = 'salesforce_opportunity_id';

    /** @var OrderFactory  */
    protected $_orderFactory;

    /** @var Account  */
    protected $_account;

    /** @var Contact  */
    protected $_contact;

    /** @var Product  */
    protected $_product;

    /** @var Order  */
    protected $_order;

    /** @var Job  */
    protected $_job;

    /** @var Data  */
    protected $_data;

    /** @var DataGetter  */
    protected $dataGetter;

    /** @var array|null */
    protected $existedOrders = null;

    /** @var array|null */
    protected $createOpportunityIds = null;

    /** @var array|null */
    protected $updateOpportunityIds = null;

    /**
     * Opportunity constructor.
     * @param OrderFactory $orderFactory
     * @param Order $order
     * @param Account $account
     * @param Contact $contact
     * @param Product $product
     * @param Data $data
     * @param Job $job
     * @param DataGetter $dataGetter
     * @param ReportCollectionFactory $reportCollectionFactory
     * @param \Magenest\Salesforce\Helper\Data $helperData
     * @param Json $json
     * @param Serialize $serialize
     * @param ScopeConfigInterface $scopeConfig
     * @param ResourceModelConfig $resourceConfig
     * @param ReportFactory $reportFactory
     * @param QueueCollectionFactory $queueCollectionFactory
     * @param QueueResource $queueResource
     * @param RequestLogFactory $requestLogFactory
     * @param ClientFactory $client
     * @param Iterator $iterator
     * @param Logger $logger
     */
    public function __construct(
        OrderFactory $orderFactory,
        Order $order,
        Account $account,
        Contact $contact,
        Product $product,
        Data $data,
        Job $job,
        DataGetter $dataGetter,
        ReportCollectionFactory $reportCollectionFactory,
        \Magenest\Salesforce\Helper\Data $helperData,
        Json $json,
        Serialize $serialize,
        ScopeConfigInterface $scopeConfig,
        ResourceModelConfig $resourceConfig,
        ReportFactory $reportFactory,
        QueueCollectionFactory $queueCollectionFactory,
        QueueResource $queueResource,
        RequestLogFactory $requestLogFactory,
        ClientFactory $client,
        Iterator $iterator,
        Logger $logger
    ) {
        $this->_orderFactory = $orderFactory;
        $this->_order        = $order;
        $this->_account      = $account;
        $this->_contact      = $contact;
        $this->_product      = $product;
        $this->_data         = $data;
        $this->_type         = 'Opportunity';
        $this->_table        = 'opportunity';
        $this->_job          = $job;
        $this->dataGetter    = $dataGetter;
        parent::__construct(
            $reportCollectionFactory,
            $helperData,
            $json,
            $serialize,
            $scopeConfig,
            $resourceConfig,
            $reportFactory,
            $queueCollectionFactory,
            $queueResource,
            $requestLogFactory,
            $iterator,
            $client,
            $logger
        );
    }

    /**
     * @param $incrementId
     * @return bool|mixed|string
     * @throws LocalizedException
     */
    public function sync($incrementId)
    {
        /** @var OrderModel $order */
        $order = $this->_orderFactory->create()->loadByIncrementId($incrementId);
        $date  = date('Y-m-d', strtotime($order->getCreatedAt()));

        $params = $this->_data->getOpportunity($order, $this->_type);

        $params  += [
            'CloseDate' => $date,
            'StageName' => 'Prospecting',
        ];
        $params  = array_replace($params, ['Name' => $incrementId]);
        $existed = $this->checkSyncOpportunity($order);
        if (!$existed) {
            $this->_helperData->saveLog("Params mapping: " . $this->_json->serialize($params));
            $opportunityId = $this->createRecords($this->_type, $params, $incrementId);
            $this->saveAttribute($order, $opportunityId);
        } else {
            $opportunityId = $existed['sid'];
            $this->updateRecords($this->_type, $opportunityId, $params, $incrementId);
        }

        /**
         * Sync OpportunityLineItem
         */
        $pricebook2Id = $this->_getStandardPriceBookId();
        $dataRaw = $this->getOpportunityItemsData($order, $pricebook2Id, !$existed);
        $params = $dataRaw['params'];
        $itemIds = $dataRaw['itemIds'];

        $this->_helperData->saveLog("OpportunityLineItem params mapping: " . $this->_json->serialize($params));
        $response = $this->_job->sendBatchRequest('insert', 'OpportunityLineItem', $this->_json->serialize($params));
        $this->saveReports('create', 'OpportunityLineItem', $response, $itemIds, $params);

        return $opportunityId;
    }

    /**
     * @param \Magento\Sales\Model\Order\Item $item
     * @param $pricebook2Id
     * @param $opportunityId
     * @return array
     * @throws LocalizedException
     */
    private function getOpportunityItemData(\Magento\Sales\Model\Order\Item $item, $pricebook2Id, $opportunityId)
    {
        $info = [];
        if (!empty($opportunityId)) {
            $product = $item->getProduct();
            $qty     = $item->getQtyOrdered();
            $price   = $item->getPrice();
            if ($price > 0) {
                $productId = $this->_product->syncOrderItem($item);
                $pricebookEntryId = $this->getPricebookEntryId(
                    $productId,
                    $pricebook2Id,
                    $product ? ($product->getPrice() ?: 0) : ($item->getOriginalPrice() ?: 0),
                    $item->getProductId()
                );
                if ($productId && $pricebookEntryId) {
                    if (!$this->searchOrderItems('OpportunityLineItem', 'OpportunityId', $opportunityId, $pricebookEntryId, (double)$qty, (double)$price)) {
                        $info = [
                            'PricebookEntryId' => $pricebookEntryId,
                            'OpportunityId' => $opportunityId,
                            'Quantity' => $qty,
                            'UnitPrice' => $price
                        ];
                    }
                }
            }
        }

        return $info;
    }

    /**
     * @param \Magento\Sales\Model\Order $order
     * @param $pricebook2Id
     * @param $syncSide
     * @return array[]
     * @throws LocalizedException
     */
    private function getOpportunityItemsData(\Magento\Sales\Model\Order $order, $pricebook2Id, $syncSide)
    {
        $params = [];
        $itemIds = [];
        $opportunityId = $order->getData(self::SALESFORCE_OPPORTUNITY_ATTRIBUTE_CODE);
        foreach ($order->getAllVisibleItems() as $item) {
            $data = $this->getOpportunityItemData($item, $pricebook2Id, $opportunityId);
            if (!empty($data)) {
                $params[]  = $data;
                $itemIds[] = ['mid' => $item->getProductId()];
            }
        }

        if ($syncSide) {
            if ($taxInfo = $this->_order->getTaxItemInfo($order, $opportunityId)) {
                $taxInfo['OpportunityId'] = $opportunityId;
                unset($taxInfo['OrderId']);
                $params[]  = $taxInfo;
                $itemIds[] = ['mid' => 'TAX'];
            }
            if ($shippingInfo = $this->_order->getShippingItemInfo($order, $opportunityId)) {
                $shippingInfo['OpportunityId'] = $opportunityId;
                unset($shippingInfo['OrderId']);
                $params[]  = $shippingInfo;
                $itemIds[] = ['mid' => 'SHIPPING'];
            }
            if ($discountInfo = $this->_order->getDiscountItemInfo($order, $opportunityId)) {
                $discountInfo['OpportunityId'] = $opportunityId;
                unset($discountInfo['OrderId']);
                $params[]  = $discountInfo;
                $itemIds[] = ['mid' => 'DISCOUNT'];
            }
        }

        return [
            'params' => $params,
            'itemIds' => $itemIds
        ];
    }

    /**
     * @return array|mixed|string|null
     */
    public function syncAllOpportunities()
    {
        try {
            $orders      = $this->_orderFactory->create()->getCollection();
            $lastOrderId = $orders->getLastItem()->getId();
            $count       = 0;
            $response    = [];
            /** @var OrderModel $order */
            foreach ($orders as $order) {
                $this->addRecord($order->getIncrementId());
                $count++;
                if ($count >= 10000 || $order->getId() == $lastOrderId) {
                    $response += $this->syncQueue();
                }
            }
            return $response;
        } catch (\Throwable $e) {
            $this->_logger->debug($e->getMessage());
        }
        return null;
    }

    /**
     * @param $orderIncrementId
     * @throws LocalizedException
     */
    public function addRecord($orderIncrementId)
    {
        $order = $this->_orderFactory->create()->loadByIncrementId($orderIncrementId);
        $id    = $this->checkSyncOpportunity($order);
        if ($order->getIncrementId()) {
            if (!$id) {
                $this->addToCreateOpportunityQueue($order);
            } else {
                $this->addToUpdateOpportunityQueue($id['mObj'], $id['sid']);
            }
        } else {
            $this->saveReport(null, 'insert', $this->_type, 2, __("The Order with this increment id failed to load."), $orderIncrementId);
        }
    }

    /**
     * @param $order
     * @return array|false
     * @throws LocalizedException
     */
    public function checkSyncOpportunity($order)
    {
        $orderIncrementId = $order->getIncrementId();
        $opportunityInReport = $this->checkLogReportByMagentoId($orderIncrementId);
        if ($opportunityInReport->getData('record_id')) {
            return [
                'mObj' => $order,
                'sid'  => $opportunityInReport->getRecordId()
            ];
        }
        $opportunityId = $this->searchRecords($this->_type, 'Name', $orderIncrementId);
        if ($opportunityId) {
            return [
                'mObj' => $order,
                'sid'  => $opportunityId
            ];
        }
        return false;
    }

    /**
     * @param OrderModel $order
     *
     * @return array|bool
     * @throws \Exception
     */
    public function checkExistedOpportunities($order)
    {
        $existedOrders    = $this->getAllSalesforceOpportunity();
        $orderIncrementId = $order->getIncrementId();
        if (isset($existedOrders[$orderIncrementId])) {
            unset($this->existedOrders[$orderIncrementId]);
            return [
                'mObj' => $order,
                'sid'  => $existedOrders[$orderIncrementId]['Id']
            ];
        }
        return false;
    }

    /**
     * @return array|null
     * @throws \Exception
     */
    public function getAllSalesforceOpportunity()
    {
        if ($this->existedOrders !== null) {
            return $this->existedOrders;
        }
        $existedOrders = $this->dataGetter->getAllSalesforceOpportunities();
        $allOrders     = [];
        foreach ($existedOrders as $key => $value) {
            $allOrders[$value['Name']] = $value;
        }
        $this->existedOrders = $allOrders;
        return $this->existedOrders;
    }

    /**
     * @return array|mixed|string|null
     * @throws LocalizedException
     */
    public function syncQueue()
    {
        $createOpportunityResponse = $this->createOpportunities();
        $this->saveAttributes($this->createOpportunityIds, $createOpportunityResponse);
        $updateOpportunityResponse         = $this->updateOpportunities();
        $createOpportunityLineItemResponse = $this->createOpportunityLineItem();
        $this->saveAttributes($this->updateOpportunityIds, $updateOpportunityResponse);

        $response = $createOpportunityResponse + $createOpportunityLineItemResponse + $updateOpportunityResponse;

        $this->unsetCreateOpportunityQueue();
        $this->unsetUpdateOpportunityQueue();
        return $response;
    }

    /**
     * @param OrderModel $order
     */
    protected function addToCreateOpportunityQueue($order)
    {
        $this->createOpportunityIds[] = [
            'mid'  => $order->getIncrementId(),
            'mObj' => $order
        ];
    }

    protected function unsetCreateOpportunityQueue()
    {
        $this->createOpportunityIds = null;
    }

    /**
     * @param OrderModel $order
     * @param $salesforceId
     */
    protected function addToUpdateOpportunityQueue($order, $salesforceId)
    {
        $this->updateOpportunityIds[] = [
            'mObj' => $order,
            'mid'  => $order->getIncrementId(),
            'sid'  => $salesforceId
        ];
    }

    protected function unsetUpdateOpportunityQueue()
    {
        $this->updateOpportunityIds = null;
    }

    /**
     * @return array|mixed|string
     * @throws LocalizedException
     */
    protected function createOpportunities()
    {
        $response = [];
        if ($this->createOpportunityIds !== null) {
            $response = $this->sendOpportunitiesRequest($this->createOpportunityIds, 'insert');
        }
        return $response;
    }

    /**
     * @return array|mixed|string
     * @throws LocalizedException
     */
    protected function updateOpportunities()
    {
        $response = [];
        if ($this->updateOpportunityIds !== null) {
            $response = $this->sendOpportunitiesRequest($this->updateOpportunityIds, 'update');
        }
        return $response;
    }

    /**
     * @return array|mixed|string
     * @throws LocalizedException
     * @throws \Exception
     */
    protected function createOpportunityLineItem()
    {
        if (empty($this->createOpportunityIds) && empty($this->updateOpportunityIds)) {
            return [];
        }
        $params   = [];
        $itemIds  = [];
        $orderIds = ($this->createOpportunityIds ?? []) + ($this->updateOpportunityIds ?? []);
        $pricebook2Id = $this->_getStandardPriceBookId();
        /** @var OrderModel $order */
        foreach ($orderIds as $orderId) {
            if ($this->checkSyncOpportunity($orderId['mObj'])) {
                $dataRaw = $this->getOpportunityItemsData(
                    $orderId['mObj'],
                    $pricebook2Id,
                    in_array($orderId, $this->createOpportunityIds ?? [])
                );
                $params = array_merge($params, $dataRaw['params']);
                $itemIds = array_merge($itemIds, $dataRaw['itemIds']);
            }
        }
        $response = $this->_job->sendBatchRequest('insert', 'OpportunityLineItem', $this->_json->serialize($params));
        $this->saveReports('create', 'OpportunityLineItem', $response, $itemIds, $params);
        return $response;
    }

    /**
     * @param $opportunityIds
     * @param $operation
     * @return mixed|string
     * @throws LocalizedException
     */
    protected function sendOpportunitiesRequest($opportunityIds, $operation)
    {
        $params = [];
        foreach ($opportunityIds as $orderId) {
            $order = $orderId['mObj'];
            $date  = date('Y-m-d', strtotime($order->getCreatedAt()));
            $info  = $this->_data->getOpportunity($order, $this->_type);
            $info  += [
                'CloseDate' => $date,
                'StageName' => 'Prospecting',
            ];
            $info  = array_replace($info, ['Name' => $order->getIncrementId()]);
            if (isset($orderId['sid'])) {
                $info += ['Id' => $orderId['sid']];
            }
            $params[] = $info;
        }
        $response = $this->_job->sendBatchRequest($operation, $this->_type, $this->_json->serialize($params));
        $this->saveReports($operation, $this->_type, $response, $opportunityIds, $params);
        return $response;
    }

    /**
     * @param $orderIds
     * @param $response
     * @throws \Exception
     */
    protected function saveAttributes($orderIds, $response)
    {
        if (empty($orderIds)) {
            return;
        }
        if (is_array($response) && is_array($orderIds)) {
            $total = count($response);
            for ($i = 0; $i < $total; $i++) {
                if (isset($orderIds[$i])) {
                    $order = $orderIds[$i]['mObj'];
                    if (isset($response[$i]['id']) && $order->getId()) {
                        $this->saveAttribute($order, $response[$i]['id']);
                    }
                }
            }
        }
    }

    /**
     * @param OrderModel $order
     * @param $salesforceId
     *
     * @throws \Exception
     */
    protected function saveAttribute($order, $salesforceId)
    {
        try {
            $resource = $order->getResource();
            $order->setData(self::SALESFORCE_OPPORTUNITY_ATTRIBUTE_CODE, $salesforceId);
            $resource->saveAttribute($order, self::SALESFORCE_OPPORTUNITY_ATTRIBUTE_CODE);
        } catch (\Throwable $e) {
            $this->_logger->critical($e->getMessage(), ['trace' => $e->getTraceAsString()]);
        }
    }
}
